@@ -1,15 +1,19 @@
 BUILD_BIT='64bit'
 BUILD_CPU='unknown'
-BUILD_DATE='2020-12-04'
+BUILD_DATE='2020-12-23'
 BUILD_MACHINE='aarch64'
 BUILD_OS='Linux unknown unknown unknown'
-BUILD_TIME='23:49'
+BUILD_TIME='03:21'
 COMPILER_VERSION='aarch64-unknown-linux-gnu-gcc (GCC) 10.2.0'
 CYTHON_VERSION='0.29.21'
 LINKER_VERSION='GNU ld (GNU Binutils) 2.35'
-PYTHON_VERSION='3.9.0'
+PYTHON_VERSION='3.9.1'
 RELEASE_BUILD=True
-lib_gtk='3.24.23'
+lib_gtk='3.24.24'
+lib_gtk_x11='3.24.24'
+lib_py3cairo='1.20.0'
+lib_pygobject='3.38.0'
+lib_python3='3.9'
 lib_vpx='1.9.0'
 lib_x11='1.7.0'
 lib_x264='0.160.3011M'
