@@ -1,5 +1,5 @@
 # This file is part of Xpra.
-# Copyright (C) 2010-2018 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2010-2020 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
@@ -15,6 +15,8 @@ avsynclog = Logger("av-sync")
 log = Logger("client", "sound")
 
 AV_SYNC_DELTA = envint("XPRA_AV_SYNC_DELTA")
+DELTA_THRESHOLD = envint("XPRA_AV_SYNC_DELTA_THRESHOLD", 40)
+DEFAULT_AV_SYNC_DELAY = envint("XPRA_DEFAULT_AV_SYNC_DELAY", 150)
 
 
 """
@@ -38,9 +40,10 @@ class AudioClient(StubClientMixin):
         #sound state:
         self.on_sink_ready = None
         self.sound_sink = None
-        self.min_sound_sequence = 0
+        self.sound_sink_sequence = 0
         self.server_sound_eos_sequence = False
         self.sound_source = None
+        self.sound_source_sequence = 0
         self.sound_in_bytecount = 0
         self.sound_out_bytecount = 0
         self.server_av_sync = False
@@ -51,10 +54,11 @@ class AudioClient(StubClientMixin):
         self.server_sound_receive = False
         self.server_sound_send = False
         self.server_sound_bundle_metadata = False
-        self.server_ogg_latency_fix = False
         self.queue_used_sent = None
+        #duplicated from ServerInfo mixin:
+        self._remote_machine_id = None
 
-    def init(self, opts, _extra_args=()):
+    def init(self, opts):
         self.av_sync = opts.av_sync
         self.sound_properties = typedict()
         self.speaker_allowed = sound_option(opts.speaker) in ("on", "off")
@@ -84,7 +88,7 @@ class AudioClient(StubClientMixin):
                     self.sound_properties = query_sound()
                     assert self.sound_properties, "query did not return any data"
                     def vinfo(k):
-                        val = self.sound_properties.strlistget(k)
+                        val = self.sound_properties.strtupleget(k)
                         assert val, "%s not found in sound properties" % k
                         return ".".join(val[:3])
                     bits = self.sound_properties.intget("python.bits", 32)
@@ -96,8 +100,8 @@ class AudioClient(StubClientMixin):
                     log.error(" %s", e)
                     self.speaker_allowed = False
                     self.microphone_allowed = False
-        encoders = self.sound_properties.strlistget("encoders", [])
-        decoders = self.sound_properties.strlistget("decoders", [])
+        encoders = self.sound_properties.strtupleget("encoders")
+        decoders = self.sound_properties.strtupleget("decoders")
         self.speaker_codecs = sound_option_or_all("speaker-codec", opts.speaker_codec, decoders)
         self.microphone_codecs = sound_option_or_all("microphone-codec", opts.microphone_codec, encoders)
         if not self.speaker_codecs:
@@ -106,7 +110,6 @@ class AudioClient(StubClientMixin):
             self.microphone_allowed = False
         self.speaker_enabled = self.speaker_allowed and sound_option(opts.speaker)=="on"
         self.microphone_enabled = self.microphone_allowed and opts.microphone.lower()=="on"
-        self.av_sync = opts.av_sync
         log("speaker: codecs=%s, allowed=%s, enabled=%s", encoders, self.speaker_allowed, csv(self.speaker_codecs))
         log("microphone: codecs=%s, allowed=%s, enabled=%s, default device=%s",
             decoders, self.microphone_allowed, csv(self.microphone_codecs), self.microphone_device)
@@ -136,7 +139,14 @@ class AudioClient(StubClientMixin):
         if self.sound_sink:
             self.stop_receiving_sound()
 
-    def get_audio_capabilities(self):
+
+    def get_caps(self) -> dict:
+        d = {}
+        updict(d, "av-sync", self.get_avsync_capabilities())
+        updict(d, "sound", self.get_audio_capabilities())
+        return d
+
+    def get_audio_capabilities(self) -> dict:
         if not self.sound_properties:
             return {}
         #we don't know if the server supports new codec names,
@@ -152,25 +162,25 @@ class AudioClient(StubClientMixin):
         log("audio capabilities: %s", caps)
         return caps
 
-    def get_caps(self):
-        return updict({}, "sound", self.get_audio_capabilities())
+    def get_avsync_capabilities(self) -> dict:
+        if not self.av_sync:
+            return {}
+        return {
+            ""              : True,
+            "delay.default" : max(0, DEFAULT_AV_SYNC_DELAY + AV_SYNC_DELTA),
+            }
 
 
-    def setup_connection(self, _conn):
-        pass
-
-    def parse_server_capabilities(self):
-        c = self.server_capabilities
+    def parse_server_capabilities(self, c : typedict) -> bool:
         self.server_av_sync = c.boolget("av-sync.enabled")
         avsynclog("av-sync: server=%s, client=%s", self.server_av_sync, self.av_sync)
         self.server_pulseaudio_id = c.strget("sound.pulseaudio.id")
         self.server_pulseaudio_server = c.strget("sound.pulseaudio.server")
-        self.server_sound_decoders = c.strlistget("sound.decoders", ())
-        self.server_sound_encoders = c.strlistget("sound.encoders", ())
+        self.server_sound_decoders = c.strtupleget("sound.decoders")
+        self.server_sound_encoders = c.strtupleget("sound.encoders")
         self.server_sound_receive = c.boolget("sound.receive")
         self.server_sound_send = c.boolget("sound.send")
         self.server_sound_bundle_metadata = c.boolget("sound.bundle-metadata")
-        self.server_ogg_latency_fix = c.boolget("sound.ogg-latency-fix", False)
         log("pulseaudio id=%s, server=%s, sound decoders=%s, sound encoders=%s, receive=%s, send=%s",
                  self.server_pulseaudio_id, self.server_pulseaudio_server,
                  csv(self.server_sound_decoders), csv(self.server_sound_encoders),
@@ -292,6 +302,7 @@ class AudioClient(StubClientMixin):
             if not ss:
                 return False
             self.sound_source = ss
+            ss.sequence = self.sound_source_sequence
             ss.connect("new-buffer", self.new_sound_buffer)
             ss.connect("state-changed", sound_source_state_changed)
             ss.connect("new-stream", self.new_stream)
@@ -330,7 +341,11 @@ class AudioClient(StubClientMixin):
             log.warn("Warning: cannot stop sound source which has not been started")
             return
         #tell the server to stop:
-        self.send("sound-data", ss.codec or "", "", {"end-of-stream" : True})
+        self.send("sound-data", ss.codec or "", "", {
+            "end-of-stream" : True,
+            "sequence"      : ss.sequence,
+            })
+        self.sound_source_sequence += 1
         ss.cleanup()
 
     def start_receiving_sound(self):
@@ -354,9 +369,6 @@ class AudioClient(StubClientMixin):
                 self.no_matching_codec_error("speaker", self.server_sound_encoders, self.speaker_codecs)
                 return
             codec = matching_codecs[0]
-            if not self.server_ogg_latency_fix and codec in ("flac", "opus", "speex"):
-                log.warn("Warning: this server's sound support is out of date")
-                log.warn(" the sound latency with the %s codec will be high", codec)
             def sink_ready(*args):
                 scodec = codec
                 log("sink_ready(%s) codec=%s (server codec name=%s)", args, codec, scodec)
@@ -377,12 +389,12 @@ class AudioClient(StubClientMixin):
         if self.speaker_enabled:
             self.speaker_enabled = False
             self.emit("speaker-changed")
-        if tell_server:
-            self.send("sound-control", "stop", self.min_sound_sequence)
-        self.min_sound_sequence += 1
-        self.send("sound-control", "new-sequence", self.min_sound_sequence)
-        if ss is None:
+        if not ss:
             return
+        if tell_server and ss.sequence==self.sound_sink_sequence:
+            self.send("sound-control", "stop", self.sound_sink_sequence)
+        self.sound_sink_sequence += 1
+        self.send("sound-control", "new-sequence", self.sound_sink_sequence)
         self.sound_sink = None
         log("stop_receiving_sound(%s) calling %s", tell_server, ss.cleanup)
         ss.cleanup()
@@ -405,15 +417,17 @@ class AudioClient(StubClientMixin):
         #not shown in the UI, so don't bother with emitting a signal:
         #self.emit("speaker-changed")
     def sound_sink_error(self, sound_sink, error):
+        log("sound_sink_error(%s, %s) exit_code=%s, current sink=%s", sound_sink, error, self.exit_code, self.sound_sink)
         if self.exit_code is not None:
             #exiting
             return
         if sound_sink!=self.sound_sink:
             log("sound_sink_error(%s, %s) not the current sink, ignoring it", sound_sink, error)
             return
-        self.may_notify_audio("Speaker forwarding error", error)
+        estr = bytestostr(error).replace("gst-resource-error-quark: ", "")
+        self.may_notify_audio("Speaker forwarding error", estr)
         log.warn("Error: stopping speaker:")
-        log.warn(" %s", bytestostr(error).replace("gst-resource-error-quark: ", ""))
+        log.warn(" %s", estr)
         self.stop_receiving_sound()
     def sound_process_stopped(self, sound_sink, *args):
         if self.exit_code is not None:
@@ -450,6 +464,7 @@ class AudioClient(StubClientMixin):
             ss = start_receiving_sound(codec)
             if not ss:
                 return False
+            ss.sequence = self.sound_sink_sequence
             self.sound_sink = ss
             ss.connect("state-changed", self.sound_sink_state_changed)
             ss.connect("error", self.sound_sink_error)
@@ -466,23 +481,26 @@ class AudioClient(StubClientMixin):
 
     def new_sound_buffer(self, sound_source, data, metadata, packet_metadata=()):
         log("new_sound_buffer(%s, %s, %s, %s)", sound_source, len(data or ()), metadata, packet_metadata)
-        if not self.sound_source:
+        if sound_source.sequence<self.sound_source_sequence:
+            log("sound buffer dropped: old sequence number: %s (current is %s)",
+                sound_source.sequence, self.sound_source_sequence)
             return
         self.sound_out_bytecount += len(data)
         for x in packet_metadata:
             self.sound_out_bytecount += len(x)
+        metadata["sequence"] = sound_source.sequence
         if packet_metadata:
             if not self.server_sound_bundle_metadata:
                 #server does not support bundling, send packet metadata as individual packets before the main packet:
                 for x in packet_metadata:
-                    self.send_sound_data(sound_source, x)
+                    self.send_sound_data(sound_source, x, metadata)
                 packet_metadata = ()
             else:
                 #the packet metadata is compressed already:
                 packet_metadata = Compressed("packet metadata", packet_metadata, can_inline=True)
         self.send_sound_data(sound_source, data, metadata, packet_metadata)
 
-    def send_sound_data(self, sound_source, data, metadata={}, packet_metadata=None):
+    def send_sound_data(self, sound_source, data, metadata, packet_metadata=None):
         codec = sound_source.codec
         packet_data = [codec, Compressed(codec, data), metadata]
         if packet_metadata:
@@ -504,8 +522,8 @@ class AudioClient(StubClientMixin):
             self.sound_in_bytecount += len(data)
         #verify sequence number if present:
         seq = metadata.intget("sequence", -1)
-        if self.min_sound_sequence>0 and 0<=seq<self.min_sound_sequence:
-            log("ignoring sound data with old sequence number %s (now on %s)", seq, self.min_sound_sequence)
+        if self.sound_sink_sequence>0 and 0<=seq<self.sound_sink_sequence:
+            log("ignoring sound data with old sequence number %s (now on %s)", seq, self.sound_sink_sequence)
             return
 
         if not self.speaker_enabled:
@@ -556,14 +574,14 @@ class AudioClient(StubClientMixin):
         if data or packet_metadata:
             ss.add_data(data, metadata, packet_metadata)
         if self.av_sync and self.server_av_sync:
-            info = ss.get_info()
-            queue_used = info.get("queue.cur") or info.get("queue", {}).get("cur")
+            qinfo = typedict(ss.get_info()).dictget("queue")
+            queue_used = typedict(qinfo or {}).intget("cur", None)
             if queue_used is None:
                 return
             delta = (self.queue_used_sent or 0)-queue_used
             #avsynclog("server sound sync: queue info=%s, last sent=%s, delta=%s",
             #    dict((k,v) for (k,v) in info.items() if k.startswith("queue")), self.queue_used_sent, delta)
-            if self.queue_used_sent is None or abs(delta)>=80:
+            if self.queue_used_sent is None or abs(delta)>=DELTA_THRESHOLD:
                 avsynclog("server sound sync: sending updated queue.used=%i (was %s)",
                           queue_used, (self.queue_used_sent or "unset"))
                 self.queue_used_sent = queue_used
@@ -576,6 +594,4 @@ class AudioClient(StubClientMixin):
     def init_authenticated_packet_handlers(self):
         log("init_authenticated_packet_handlers()")
         #these handlers can run directly from the network thread:
-        self.set_packet_handlers(self._packet_handlers, {
-            "sound-data":           self._process_sound_data,
-            })
+        self.add_packet_handler("sound-data", self._process_sound_data, False)
