@@ -1,40 +1,36 @@
 # This file is part of Xpra.
-# Copyright (C) 2013-2018 Antoine Martin <antoine@devloop.org.uk>
+# Copyright (C) 2013-2019 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
-#authentication from a file containing just the password
-
 import os.path
 
 from xpra.net.digest import get_salt, choose_digest
 from xpra.os_util import strtobytes
 from xpra.server.auth.sys_auth_base import SysAuthenticator
 from xpra.log import Logger
+
 log = Logger("auth")
 
 
-#legacy interface: this is to inject the "--password-file=" option
-#this is shared by all instances
-password_file = None
-def init(opts):
-    global password_file
-    password_file = opts.password_file
-
-
 class FileAuthenticatorBase(SysAuthenticator):
     def __init__(self, username, **kwargs):
-        filename = kwargs.pop("filename", password_file)
-        if filename and not os.path.isabs(filename):
+        password_file = kwargs.pop("filename", None)
+        log("FileAuthenticatorBase password_file=%s", password_file)
+        if not password_file:
+            log.warn("Warning: %r authentication module is missing the 'filename' option", self)
+            log.warn(" all authentication attempts will fail")
+        elif not os.path.isabs(password_file):
             exec_cwd = kwargs.get("exec_cwd", os.getcwd())
-            filename = os.path.join(exec_cwd, filename)
-        SysAuthenticator.__init__(self, username, **kwargs)
-        self.password_filename = filename
+            password_file = os.path.join(exec_cwd, password_file)
+        log("FileAuthenticatorBase filename=%s", password_file)
+        super().__init__(username, **kwargs)
+        self.password_filename = password_file
         self.password_filedata = None
         self.password_filetime = None
         self.authenticate = self.authenticate_hmac
 
-    def requires_challenge(self):
+    def requires_challenge(self) -> bool:
         return True
 
     def get_challenge(self, digests):
@@ -50,7 +46,7 @@ class FileAuthenticatorBase(SysAuthenticator):
             return None
         return self.salt, self.digest
 
-    def get_password(self):
+    def get_password(self) -> str:
         file_data = self.load_password_file()
         if file_data is None:
             return None
@@ -59,7 +55,7 @@ class FileAuthenticatorBase(SysAuthenticator):
     def parse_filedata(self, data):
         return data
 
-    def load_password_file(self):
+    def load_password_file(self) -> bytes:
         if not self.password_filename:
             return None
         full_path = os.path.abspath(self.password_filename)
@@ -83,7 +79,7 @@ class FileAuthenticatorBase(SysAuthenticator):
                     self.password_filedata = None
         return self.password_filedata
 
-    def stat_password_filetime(self):
+    def stat_password_filetime(self) -> int:
         try:
             full_path = os.path.abspath(self.password_filename)
             v = os.stat(full_path).st_mtime
