@@ -8,8 +8,8 @@ import os
 import sys
 import time
 from multiprocessing import Queue as MQueue, freeze_support #@UnresolvedImport
+from gi.repository import GLib
 
-from xpra.gtk_common.gobject_compat import import_glib
 from xpra.util import (
     LOGIN_TIMEOUT, AUTHENTICATION_ERROR, SESSION_NOT_FOUND, SERVER_ERROR,
     repr_ellipsized, print_nested_dict, csv, envfloat, envbool, envint, typedict,
@@ -18,7 +18,6 @@ from xpra.os_util import (
     get_username_for_uid, get_groups, get_home_for_uid, bytestostr,
     getuid, getgid, WIN32, POSIX,
     monotonic_time,
-    register_SIGUSR_signals,
     )
 from xpra.server.server_core import ServerCore
 from xpra.server.control_command import ArgsControlCommand, ControlError
@@ -32,17 +31,17 @@ from xpra.log import Logger
 log = Logger("proxy")
 authlog = Logger("proxy", "auth")
 
-glib = import_glib()
-
 freeze_support()
 
 
 PROXY_SOCKET_TIMEOUT = envfloat("XPRA_PROXY_SOCKET_TIMEOUT", "0.1")
 PROXY_WS_TIMEOUT = envfloat("XPRA_PROXY_WS_TIMEOUT", "1.0")
 assert PROXY_SOCKET_TIMEOUT>0, "invalid proxy socket timeout"
-CAN_STOP_PROXY = envbool("XPRA_CAN_STOP_PROXY", getuid()!=0)
+CAN_STOP_PROXY = envbool("XPRA_CAN_STOP_PROXY", getuid()!=0 or WIN32)
 STOP_PROXY_SOCKET_TYPES = os.environ.get("XPRA_STOP_PROXY_SOCKET_TYPES", "unix-domain,named-pipe").split(",")
-PROXY_INSTANCE_THREADED = envbool("XPRA_PROXY_INSTANCE_THREADED", False)
+STOP_PROXY_AUTH_SOCKET_TYPES = os.environ.get("XPRA_STOP_PROXY_AUTH_SOCKET_TYPES", "unix-domain").split(",")
+#something (a thread lock?) doesn't allow us to use multiprocessing on MS Windows:
+PROXY_INSTANCE_THREADED = envbool("XPRA_PROXY_INSTANCE_THREADED", WIN32)
 PROXY_CLEANUP_GRACE_PERIOD = envfloat("XPRA_PROXY_CLEANUP_GRACE_PERIOD", "0.5")
 
 MAX_CONCURRENT_CONNECTIONS = envint("XPRA_PROXY_MAX_CONCURRENT_CONNECTIONS", 200)
@@ -88,19 +87,18 @@ class ProxyServer(ServerCore):
         self.instances = {}
         #connections used exclusively for requests:
         self._requests = set()
-        self.idle_add = glib.idle_add
-        self.timeout_add = glib.timeout_add
-        self.source_remove = glib.source_remove
+        self.idle_add = GLib.idle_add
+        self.timeout_add = GLib.timeout_add
+        self.source_remove = GLib.source_remove
         self._socket_timeout = PROXY_SOCKET_TIMEOUT
         self._ws_timeout = PROXY_WS_TIMEOUT
-        register_SIGUSR_signals(glib.idle_add)
 
     def init(self, opts):
         log("ProxyServer.init(%s)", opts)
         self.pings = int(opts.pings)
         self.video_encoders = opts.proxy_video_encoders
         self._start_sessions = opts.proxy_start_sessions
-        ServerCore.init(self, opts)
+        super().init(opts)
         #ensure we cache the platform info before intercepting SIGCHLD
         #as this will cause a fork and SIGCHLD to be emitted:
         from xpra.version_util import get_platform_info
@@ -114,8 +112,9 @@ class ProxyServer(ServerCore):
 
 
     def install_signal_handlers(self, callback):
-        from xpra.gtk_common.gobject_compat import register_os_signals
-        register_os_signals(callback)
+        from xpra.gtk_common.gobject_compat import register_os_signals, register_SIGUSR_signals
+        register_os_signals(callback, "Proxy Server")
+        register_SIGUSR_signals("Proxy Server")
 
 
     def make_dbus_server(self):
@@ -130,7 +129,7 @@ class ProxyServer(ServerCore):
 
     def _process_proxy_shutdown_server(self, proto, _packet):
         assert proto in self._requests
-        self.quit(False)
+        self.clean_quit(False)
 
 
     def print_screen_info(self):
@@ -144,7 +143,7 @@ class ProxyServer(ServerCore):
         pass
 
     def do_run(self):
-        self.main_loop = glib.MainLoop()
+        self.main_loop = GLib.MainLoop()
         self.main_loop.run()
 
     def handle_stop_command(self, *args):
@@ -178,11 +177,15 @@ class ProxyServer(ServerCore):
         log("stop_proxy(%s) %s", instance, v)
         #different ways of stopping for process vs threaded implementations:
         if isprocess:
+            #send message:
             if force:
                 instance.terminate()
             else:
-                #send message:
-                mq.put("stop")
+                mq.put_nowait("stop")
+                try:
+                    mq.close()
+                except Exception as e:
+                    log("%s() %s", mq.close, e)
         else:
             #direct method call:
             instance.stop(None, "proxy server request")
@@ -209,6 +212,9 @@ class ProxyServer(ServerCore):
     def do_quit(self):
         self.main_loop.quit()
         log.info("Proxy Server process ended")
+        #from now on, we can't rely on the main loop:
+        from xpra.os_util import register_SIGUSR_signals
+        register_SIGUSR_signals()
 
 
     def verify_connection_accepted(self, protocol):
@@ -218,7 +224,7 @@ class ProxyServer(ServerCore):
             self.send_disconnect(protocol, LOGIN_TIMEOUT)
 
     def hello_oked(self, proto, packet, c, auth_caps):
-        if ServerCore.hello_oked(self, proto, packet, c, auth_caps):
+        if super().hello_oked(proto, packet, c, auth_caps):
             #already handled in superclass
             return
         self.accept_client(proto, c)
@@ -244,7 +250,7 @@ class ProxyServer(ServerCore):
                 self.send_disconnect(proto, msg)
                 return
             #connection must be authenticated:
-            if not proto.authenticators:
+            if socktype in STOP_PROXY_AUTH_SOCKET_TYPES and not proto.authenticators:
                 msg = "cannot stop proxy server from unauthenticated connections"
                 log.warn("Warning: %s", msg)
                 self.send_disconnect(proto, msg)
@@ -280,6 +286,7 @@ class ProxyServer(ServerCore):
             disconnect(SESSION_NOT_FOUND, "no sessions found")
             return
         sessions = None
+        authenticator = None
         for authenticator in client_proto.authenticators:
             try:
                 auth_sessions = authenticator.get_sessions()
@@ -315,6 +322,7 @@ class ProxyServer(ServerCore):
                 uid = getuid()
                 gid = getgid()
             username = get_username_for_uid(uid)
+            password = None
             groups = get_groups(username)
             log("username(%i)=%s, groups=%s", uid, username, groups)
         else:
@@ -322,6 +330,7 @@ class ProxyServer(ServerCore):
             assert client_proto.authenticators
             for authenticator in client_proto.authenticators:
                 username = getattr(authenticator, "username", "")
+                password = authenticator.get_password()
                 if username:
                     break
         #ensure we don't loop back to the proxy:
@@ -342,8 +351,8 @@ class ProxyServer(ServerCore):
                 disconnect(SESSION_NOT_FOUND, "no displays found")
                 return
             try:
-                proc, socket_path, display = self.start_new_session(username, uid, gid, sns, displays)
-                log("start_new_session%s=%s", (username, uid, gid, sns, displays), (proc, socket_path, display))
+                proc, socket_path, display = self.start_new_session(username, password, uid, gid, sns, displays)
+                log("start_new_session%s=%s", (username, "..", uid, gid, sns, displays), (proc, socket_path, display))
             except Exception as e:
                 log("start_server_subprocess failed", exc_info=True)
                 log.error("Error: failed to start server subprocess:")
@@ -501,8 +510,8 @@ class ProxyServer(ServerCore):
                 message_queue.put("socket-handover-complete")
         start_thread(start_proxy_process, "start_proxy(%s)" % client_proto)
 
-    def start_new_session(self, username, uid, gid, new_session_dict=None, displays=()):
-        log("start_new_session%s", (username, uid, gid, new_session_dict, displays))
+    def start_new_session(self, username, _password, uid, gid, new_session_dict=None, displays=()):
+        log("start_new_session%s", (username, "..", uid, gid, new_session_dict, displays))
         sns = typedict(new_session_dict or {})
         mode = sns.get("mode", "start")
         assert mode in ("start", "start-desktop", "shadow"), "invalid start-new-session mode '%s'" % mode
@@ -595,7 +604,7 @@ class ProxyServer(ServerCore):
 
 
     def get_info(self, proto, *_args):
-        info = ServerCore.get_minimal_server_info(self)
+        info = ServerCore.get_server_info(self)
         info.setdefault("server", {})["type"] = "Python/GLib/proxy"
         #only show more info if we have authenticated
         #as the user running the proxy server process:
